/* This program illustrates the capabilities of PAMR
 * to manage distributed hierarchical grids
 * It creates 2D hierarchical grid with NMAXGRIDS levels
 * The levels are centered around a Gaussian.
 * The program then `evolves' the Gaussian pulse around a circle.
 * It regrids every NMAXGRIDS steps.
 * Since there is no real evolution in this code the number of ghost 
 * zones can be set to zero (but is kept at 1)
 * Each time step the code simply redraws the gaussian centered at
 * xc[0] = x0 + arm*cos(2*PI*time)
 * xc[1] = y0 + arm*sin(2*PI*time)
 * x0,y0 is the center of the global bounding box
 *
 * Saves the grid functions after every OUTMODULO steps.
 * Also saves the grid function just after regriding (at time + dt/2).
 *
 * Try to run the code with REGRIDSTEP = 10
 * Why the regrid doesn't work?
 */
#include <stdio.h>
#include <math.h>
#include <pamr.h>

#define DIM  2 // dimensionality of the problem
#define XMIN 0.0 
#define XMAX 1.0
#define YMIN 0.0 
#define YMAX 1.0
#define RHO 2     // refinment ratio
#define NGLEV 3   // number of grid levels
#define NX   25  // total number of grid points in x-direction
#define NY   25  // total number of grid points in y-direction
#define NTLEV 1   // number of time levels
#define NMAXGRIDS 3 // maximum number of grids (1 for unigrid)
#define NMAXGF 1 // maximum number of grid functions
#define OUTMODULO 1 // save every OUTMODULO steps
#define NSTEPS 101 // number of evolution steps
#define REGRIDSTEP 5 // regriding period
#define NGHOSTS 1 // number of ghost zones
#define HFRAC 0.25 // defines the size of the grids in hierarchy

#define GF_F 0 // symbolic name of index for grid function "f"
#define max(a, b)  (((a) > (b)) ? (a) : (b))
#define min(a, b)  (((a) < (b)) ? (a) : (b))


// declare the fortran functions - note the trailing underscore
void initialize_(int *Nx,int *Ny,real *x,real *y,real *f,real *xc,real *delta);
void calc_center(real *time,real *arm,real *x0,real *y0,real *xc);

// define global variables
real *f[NTLEV];
int f_gfn[NTLEV];

real *x,*y;
real xmin,xmax,ymin,ymax;
int ghost_width[2*DIM],Nx,Ny,phys_bdy[2*DIM],size;
int Lf,rank;
real base_bbox[2*DIM],bbox[2*DIM],dx,dy,lambda;
int base_shape[DIM];

// set the gfn's for all the variables
void set_gfns(void)
{
  int lev;
  for(lev=1;lev<=NTLEV;lev++) {
    if ((f_gfn[lev-1]=PAMR_get_gfn("f",PAMR_AMRH,lev))<0) perror("set_gnfs error");
  }
}

// load pointers
void ldptr(void)
{
   int dim,ngfs,lev,shape[DIM];
   real t,*x0[DIM],*gfs[NMAXGF*NTLEV],dx0[DIM];
   static int first=1;

   if (first) // call only once
   {
      first=0;
      set_gfns();
   }
   if (!(PAMR_get_g_attribs(&rank,&dim,shape,bbox,ghost_width,&t,&ngfs,x0,gfs)))
      perror("ldptr: PAMR_get_g_attribs failed");

   for(lev=0;lev<NTLEV;lev++) {
     f[lev]=gfs[f_gfn[lev]-1];
   }

   x=x0[0]; dx=x[1]-x[0];
   y=x0[1]; dy=y[1]-y[0];

   if ((bbox[0]-base_bbox[0])<dx/2) phys_bdy[0]=1; else phys_bdy[0]=0;
   if ((base_bbox[1]-bbox[1])<dx/2) phys_bdy[1]=1; else phys_bdy[1]=0;
   if ((bbox[2]-base_bbox[2])<dy/2) phys_bdy[2]=1; else phys_bdy[2]=0;
   if ((base_bbox[3]-bbox[3])<dy/2) phys_bdy[3]=1; else phys_bdy[3]=0;

   Nx=shape[0];
   Ny=shape[1];

   size=Nx*Ny;
}

// update the hierarchy bbox around the center defined by (xc[0],xc[1])
void update_gbox(int ngl,real *gbox,real *xc) {
  real width[2*DIM],pdx[2*DIM],temp,factor;
  int i,l;

  for(i=0;i<DIM;i++) {
    width[i] = base_bbox[2*i+1]-base_bbox[2*i]; // width of the global bounding box
    pdx[i] = width[i]/(base_shape[i]-1); // mesh spacing of the parent grid (to be used 
                                         // in the next loop)
    gbox[2*i] = base_bbox[2*i];          // initialize base grid
    gbox[2*i+1] = base_bbox[2*i+1];
  }
  factor = 0.5*HFRAC; // relative half width of the bounding box 
  for(l=1;l<ngl;l++) { // loop over levels (starting from second level)
    for(i=0;i<DIM;i++) { // loop over dimensions
      temp = max(base_bbox[2*i],xc[i]-width[i]*factor);
      // align with the parent grid
      gbox[2*DIM*l+2*i] = base_bbox[2*i]+(int)((temp-base_bbox[2*i])/pdx[i]+0.5)*pdx[i];
      temp = min(base_bbox[2*i+1],xc[i]+width[i]*factor); 
      // align with the parent grid
      gbox[2*DIM*l+2*i+1] = base_bbox[2*i+1]-(int)((base_bbox[2*i+1]-temp)/pdx[i]+0.5)*pdx[i];
      // adjust pdx for next level
      pdx[i] /=RHO;
    }
    factor *=HFRAC;
  }
  
}
// save variables (all levels)
void save_vars(real time) {
  char *name[NMAXGF];
  int num_save,i,l;
  num_save = 1;
  name[0] = (char *) strdup("f");
  for(l=1;l<=Lf;l++) {
    for (i=0;i<num_save;i++) {
      if (!PAMR_save_gfn(name[i],PAMR_AMRH,1,l,time,"","")) {
        perror("save_vars: error saving grid function");
      }
    }
  }
}


int main( int argc, char *argv[])
{
  int cp_rank,pamr_context;
  real q,q_size,time,dt;
  char *cp_file;
  int lev[NMAXGRIDS];
  char *varname;
  int in_amrh,in_mgh,num_tl,amr_inject,amr_interp;
  int amr_bdy_interp,amr_sync,mg_inject,mg_interp;
  int mg_sync,mg_noinj_to_amr,regrid_transfer,phys_bdy_type[2*DIM];

  int g_dim,g_shape[DIM],g_ghost_width[2*DIM],g_ngfs;

  int valid,Ng[DIM],min_lev,max_lev,num;

  int step,nsteps,outmodulo,tlev1,i,l;

  real gbox[2*DIM*NMAXGRIDS],x0,y0,xc[DIM],arm,delta;

  base_bbox[0] = XMIN;
  base_bbox[1] = XMAX;
  base_bbox[2] = YMIN;
  base_bbox[3] = YMAX;
  base_shape[0] = NX;
  base_shape[1] = NY;
  q = 0;
  q_size = 0;
  cp_file = (char *) NULL;
  cp_rank = -1;

  MPI_Init(&argc,&argv);
  pamr_context = PAMR_init_context(0,q_size,DIM,base_shape,base_bbox,cp_file,cp_rank);
  if (!pamr_context) {
    perror("Error initializing pamr_context");
    exit(-1);
  }
  // define variable attributes for unigrid (no transfers necessary)
  varname = (char *) strdup("f");
  in_amrh = 1;
  in_mgh = 0;
  num_tl = NTLEV;
  amr_inject = PAMR_STRAIGHT_INJECT;
  amr_interp = PAMR_SECOND_ORDER;
  amr_bdy_interp = PAMR_SECOND_ORDER;
  amr_sync = PAMR_SYNC;
  mg_inject = PAMR_NO_INJECT;
  mg_interp = PAMR_NO_INTERP;
  mg_sync = PAMR_NO_SYNC;
  mg_noinj_to_amr = 0;
  //regrid_transfer = PAMR_NO_TRANSFER;
  regrid_transfer = PAMR_SECOND_ORDER;

  // note that sgfn starts at 1 therefore we subtract -1 for C-style arrays
  PAMR_def_var_full(varname,in_amrh,in_mgh,num_tl,amr_inject,
              amr_interp,amr_bdy_interp,amr_sync,mg_inject,mg_interp,
              mg_sync,mg_noinj_to_amr,regrid_transfer,phys_bdy_type)-1;
  Ng[0] = NGHOSTS;
  Ng[1] = NGHOSTS;
  PAMR_set_ghost_width(Ng);
  // for unigrid we have only level = 1 grid
  min_lev = 1;
  max_lev = NGLEV;
  num = NGLEV;
  for(i=0;i<NGLEV;i++) {
    lev[i] = i+1; 
  }
  time = 0.0;
  x0 = 0.5*(XMIN+XMAX);
  y0 = 0.5*(YMIN+YMAX);
  arm = 0.25*(XMAX-XMIN);
  delta = 0.1*(XMAX-XMIN);
  // calculate the center of the gaussian xc
  calc_center_(&time,&arm,&x0,&y0,xc);
  // create hierarchy gbox centered around xc
  update_gbox(NGLEV,gbox,xc);
  // compose the hierarchy
  PAMR_compose_hierarchy(min_lev,max_lev,num,lev,gbox,time);
  // initialize all levels
  tlev1 = 0;
  Lf =  PAMR_get_max_lev(PAMR_AMRH);
  for(l=1;l<=Lf;l++) { // loop over levels
    valid = PAMR_init_s_iter(l,PAMR_AMRH,0);
    while (valid) { // loop over grids at each level
      ldptr();
      initialize_(&Nx,&Ny,x,y,f[tlev1],xc,&delta);
      valid = PAMR_next_g();
    }
  }

  dt = 0.01;
  step = 0;
  while (step<NSTEPS) { // main 'evolution' loop
    if (rank ==0) printf("step=%d\n",step);
    if ((step % OUTMODULO) == 0) {
      if (rank ==0) printf("SAVING at time=%g\n",time);
      save_vars(time);
    }

    // regrid after REGRIDSTEP steps
    if ((step % REGRIDSTEP) == 0 && step>0) {
      // recalculate the grid structure
      update_gbox(NGLEV,gbox,xc);
      // regrid
      PAMR_compose_hierarchy(min_lev,max_lev,num,lev,gbox,time);
      // save the grid functions just after regriding
      save_vars(time+0.5*dt);
    }
    step++;
    time += dt;
    // this sequence emulates evolution
    calc_center_(&time,&arm,&x0,&y0,xc);
    Lf =  PAMR_get_max_lev(PAMR_AMRH);
    for(l=1;l<=Lf;l++) {
      valid = PAMR_init_s_iter(l,PAMR_AMRH,0);
      while (valid) {
        ldptr();
        initialize_(&Nx,&Ny,x,y,f[tlev1],xc,&delta);
        valid = PAMR_next_g();
      }
    }
    // not really necessary here, but for further reference
    PAMR_sync(1,tlev1+1,PAMR_AMRH,0);
  }
  PAMR_free_context(pamr_context);
}
